
from collections import OrderedDict
from pprint import pformat
from rubikscubennnsolver.RubiksCubeNNNEvenEdges import RubiksCubeNNNEvenEdges
from rubikscubennnsolver.RubiksCube444 import RubiksCube444, solved_4x4x4
from rubikscubennnsolver.RubiksCube555 import RubiksCube555, solved_5x5x5
from rubikscubennnsolver.RubiksSide import Side, SolveError
from rubikscubennnsolver.LookupTable import LookupTable, LookupTableIDA
import logging
import math
import os
import random
import re
import subprocess
import sys

log = logging.getLogger(__name__)


moves_6x6x6 = ("U", "U'", "U2", "Uw", "Uw'", "Uw2", "3Uw", "3Uw'", "3Uw2",
               "L", "L'", "L2", "Lw", "Lw'", "Lw2", "3Lw", "3Lw'", "3Lw2",
               "F" , "F'", "F2", "Fw", "Fw'", "Fw2", "3Fw", "3Fw'", "3Fw2",
               "R" , "R'", "R2", "Rw", "Rw'", "Rw2", "3Rw", "3Rw'", "3Rw2",
               "B" , "B'", "B2", "Bw", "Bw'", "Bw2", "3Bw", "3Bw'", "3Bw2",
               "D" , "D'", "D2", "Dw", "Dw'", "Dw2", "3Dw", "3Dw'", "3Dw2")
solved_6x6x6 = 'UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUURRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'


class LookupTable666UDInnerXCenterStage(LookupTable):
    """
    Stage the inner X-centers
    24!/(8!*16!) is 735,471

    lookup-table-6x6x6-step10-UD-inner-x-centers-stage.txt
    ======================================================
    1 steps has 5 entries (0 percent, 0.00x previous step)
    2 steps has 82 entries (0 percent, 16.40x previous step)
    3 steps has 1,206 entries (0 percent, 14.71x previous step)
    4 steps has 14,116 entries (1 percent, 11.70x previous step)
    5 steps has 123,404 entries (16 percent, 8.74x previous step)
    6 steps has 422,508 entries (57 percent, 3.42x previous step)
    7 steps has 173,254 entries (23 percent, 0.41x previous step)
    8 steps has 896 entries (0 percent, 0.01x previous step)

    Total: 735,471 entries
    """

    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step10-UD-inner-x-centers-stage.txt',
            '066000000000000000000660',
            linecount=735471)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Upper
            'x', 'x', 'x', 'x',
            'x', parent_state[15], parent_state[16], 'x',
            'x', parent_state[21], parent_state[22], 'x',
            'x', 'x', 'x', 'x',

            # Left
            'x', 'x', 'x', 'x',
            'x', parent_state[51], parent_state[52], 'x',
            'x', parent_state[57], parent_state[58], 'x',
            'x', 'x', 'x', 'x',

            # Front
            'x', 'x', 'x', 'x',
            'x', parent_state[87], parent_state[88], 'x',
            'x', parent_state[93], parent_state[94], 'x',
            'x', 'x', 'x', 'x',

            # Right
            'x', 'x', 'x', 'x',
            'x', parent_state[123], parent_state[124], 'x',
            'x', parent_state[129], parent_state[130], 'x',
            'x', 'x', 'x', 'x',

            # Back
            'x', 'x', 'x', 'x',
            'x', parent_state[159], parent_state[160], 'x',
            'x', parent_state[165], parent_state[166], 'x',
            'x', 'x', 'x', 'x',

            # Down
            'x', 'x', 'x', 'x',
            'x', parent_state[195], parent_state[196], 'x',
            'x', parent_state[201], parent_state[202], 'x',
            'x', 'x', 'x', 'x'
        ]

        result = ['1' if x in ('U', 'D') else '0' for x in result]
        result = ''.join(result)

        # Convert to hex
        return self.hex_format % int(result, 2)


class LookupTable666UDObliqueEdgePairingLeftOnly(LookupTable):
    """
    lookup-table-6x6x6-step21-UD-oblique-edge-pairing-left-only.txt
    ===============================================================
    1 steps has 5 entries (0 percent, 0.00x previous step)
    2 steps has 82 entries (0 percent, 16.40x previous step)
    3 steps has 1,198 entries (0 percent, 14.61x previous step)
    4 steps has 13,818 entries (1 percent, 11.53x previous step)
    5 steps has 115,638 entries (15 percent, 8.37x previous step)
    6 steps has 399,478 entries (54 percent, 3.45x previous step)
    7 steps has 204,612 entries (27 percent, 0.51x previous step)
    8 steps has 640 entries (0 percent, 0.00x previous step)

    Total: 735,471 entries
    """

    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step21-UD-oblique-edge-pairing-left-only.txt',
            '990000000099',
            linecount=735471)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Upper
            parent_state[9], 'x',
            'x', parent_state[17],
            parent_state[20], 'x',
            'x', parent_state[28],

            # Left
            parent_state[45], 'x',
            'x', parent_state[53],
            parent_state[56], 'x',
            'x', parent_state[64],

            # Front
            parent_state[81], 'x',
            'x', parent_state[89],
            parent_state[92], 'x',
            'x', parent_state[100],

            # Right
            parent_state[117], 'x',
            'x', parent_state[125],
            parent_state[128], 'x',
            'x', parent_state[136],

            # Back
            parent_state[153], 'x',
            'x', parent_state[161],
            parent_state[164], 'x',
            'x', parent_state[172],

            # Down
            parent_state[189], 'x',
            'x', parent_state[197],
            parent_state[200], 'x',
            'x', parent_state[208],
        ]

        result = ['1' if x in ('U', 'D') else '0' for x in result]
        result = ''.join(result)

        # Convert to hex
        return self.hex_format % int(result, 2)


class LookupTable666UDObliqueEdgePairingRightOnly(LookupTable):
    """
    lookup-table-6x6x6-step22-UD-oblique-edge-pairing-right-only.txt
    ================================================================
    1 steps has 5 entries (0 percent, 0.00x previous step)
    2 steps has 82 entries (0 percent, 16.40x previous step)
    3 steps has 1,198 entries (0 percent, 14.61x previous step)
    4 steps has 13,818 entries (1 percent, 11.53x previous step)
    5 steps has 115,638 entries (15 percent, 8.37x previous step)
    6 steps has 399,478 entries (54 percent, 3.45x previous step)
    7 steps has 204,612 entries (27 percent, 0.51x previous step)
    8 steps has 640 entries (0 percent, 0.00x previous step)

    Total: 735,471 entries
    """

    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step22-UD-oblique-edge-pairing-right-only.txt',
            '660000000066',
            linecount=735471)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Upper
            'x', parent_state[10],
            parent_state[14], 'x',
            'x', parent_state[23],
            parent_state[27], 'x',

            # Left
            'x', parent_state[46],
            parent_state[50], 'x',
            'x', parent_state[59],
            parent_state[63], 'x',


            # Front
            'x', parent_state[82],
            parent_state[86], 'x',
            'x', parent_state[95],
            parent_state[99], 'x',

            # Right
            'x', parent_state[118],
            parent_state[122], 'x',
            'x', parent_state[131],
            parent_state[135], 'x',

            # Back
            'x', parent_state[154],
            parent_state[158], 'x',
            'x', parent_state[167],
            parent_state[171], 'x',

            # Down
            'x', parent_state[190],
            parent_state[194], 'x',
            'x', parent_state[203],
            parent_state[207], 'x',
        ]

        result = ['1' if x in ('U', 'D') else '0' for x in result]
        result = ''.join(result)

        # Convert to hex
        return self.hex_format % int(result, 2)


class LookupTableIDA666UDObliqueEdgePairing(LookupTableIDA):
    """
    Now pair the UD oblique edges so that we can reduce the 6x6x6 centers to a 5x5x5
    (24!/(8!*16!))^2 is 540,917,591,841 so this is too large for us to build so use
    IDA and build it 8 steps deep.

    Our prune tables will be to solve on the left or right oblique edges. Each of these
    tables are 24!/(8!*16!) or 735,471
    735471/540917591841 is 0.000 001 3596729171

    This IDA search can take a few dozen seconds

    lookup-table-6x6x6-step20-UD-oblique-edge-pairing.txt
    =====================================================
    1 steps has 5 entries (0 percent, 0.00x previous step)
    2 steps has 82 entries (0 percent, 16.40x previous step)
    3 steps has 1,434 entries (0 percent, 17.49x previous step)
    4 steps has 24,198 entries (0 percent, 16.87x previous step)
    5 steps has 405,916 entries (0 percent, 16.77x previous step)
    6 steps has 6,839,392 entries (5 percent, 16.85x previous step)
    7 steps has 116,031,874 entries (94 percent, 16.97x previous step)

    Total: 123,302,901 entries
    """

    def __init__(self, parent):
        LookupTableIDA.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step20-UD-oblique-edge-pairing.txt',
            'ff00000000ff',
            moves_6x6x6,

            ("3Rw", "3Rw'", "3Lw", "3Lw'", "3Fw", "3Fw'", "3Bw", "3Bw'", # These would break up the staged UD inner x-centers
             "3Lw", "3Lw'", "3Lw2",        # can skip these for 6x6x6 cubes
             "3Dw", "3Dw'", "3Dw2",
             "3Bw", "3Bw'", "3Bw2"),

            # prune tables
            (parent.lt_UD_oblique_edge_pairing_left_only,
             parent.lt_UD_oblique_edge_pairing_right_only),
            linecount=123302901)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Upper
            parent_state[9], parent_state[10],
            parent_state[14], parent_state[17],
            parent_state[20], parent_state[23],
            parent_state[27], parent_state[28],

            # Left
            parent_state[45], parent_state[46],
            parent_state[50], parent_state[53],
            parent_state[56], parent_state[59],
            parent_state[63], parent_state[64],

            # Front
            parent_state[81], parent_state[82],
            parent_state[86], parent_state[89],
            parent_state[92], parent_state[95],
            parent_state[99], parent_state[100],

            # Right
            parent_state[117], parent_state[118],
            parent_state[122], parent_state[125],
            parent_state[128], parent_state[131],
            parent_state[135], parent_state[136],

            # Back
            parent_state[153], parent_state[154],
            parent_state[158], parent_state[161],
            parent_state[164], parent_state[167],
            parent_state[171], parent_state[172],

            # Down
            parent_state[189], parent_state[190],
            parent_state[194], parent_state[197],
            parent_state[200], parent_state[203],
            parent_state[207], parent_state[208],
        ]

        result = ['1' if x in ('U', 'D') else '0' for x in result]
        result = ''.join(result)

        # Convert to hex
        return self.hex_format % int(result, 2)


class LookupTable666LRInnerXCenters(LookupTable):
    """
    lookup-table-6x6x6-step30-LR-inner-x-centers-stage.txt
    ======================================================
    1 steps has 3 entries (0 percent, 0.00x previous step)
    2 steps has 29 entries (0 percent, 9.67x previous step)
    3 steps has 234 entries (1 percent, 8.07x previous step)
    4 steps has 1,246 entries (9 percent, 5.32x previous step)
    5 steps has 4,466 entries (34 percent, 3.58x previous step)
    6 steps has 6,236 entries (48 percent, 1.40x previous step)
    7 steps has 656 entries (5 percent, 0.11x previous step)

    Total: 12,870 entries
    """

    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step30-LR-inner-x-centers-stage.txt',
            '0660000006600000',
            linecount=12870)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Left
            'x', 'x', 'x', 'x',
            'x', parent_state[51], parent_state[52], 'x',
            'x', parent_state[57], parent_state[58], 'x',
            'x', 'x', 'x', 'x',

            # Front
            'x', 'x', 'x', 'x',
            'x', parent_state[87], parent_state[88], 'x',
            'x', parent_state[93], parent_state[94], 'x',
            'x', 'x', 'x', 'x',

            # Right
            'x', 'x', 'x', 'x',
            'x', parent_state[123], parent_state[124], 'x',
            'x', parent_state[129], parent_state[130], 'x',
            'x', 'x', 'x', 'x',

            # Back
            'x', 'x', 'x', 'x',
            'x', parent_state[159], parent_state[160], 'x',
            'x', parent_state[165], parent_state[166], 'x',
            'x', 'x', 'x', 'x',
        ]

        result = ['1' if x in ('L', 'R') else '0' for x in result]
        result = ''.join(result)

        # Convert to hex
        return self.hex_format % int(result, 2)


class LookupTable666LRObliqueEdgePairingLeftOnly(LookupTable):
    """
    lookup-table-6x6x6-step41-LR-oblique-pairing-left-only.txt
    ==========================================================
    1 steps has 3 entries (0 percent, 0.00x previous step)
    2 steps has 29 entries (0 percent, 9.67x previous step)
    3 steps has 238 entries (1 percent, 8.21x previous step)
    4 steps has 742 entries (5 percent, 3.12x previous step)
    5 steps has 1,836 entries (14 percent, 2.47x previous step)
    6 steps has 4,405 entries (34 percent, 2.40x previous step)
    7 steps has 3,774 entries (29 percent, 0.86x previous step)
    8 steps has 1,721 entries (13 percent, 0.46x previous step)
    9 steps has 122 entries (0 percent, 0.07x previous step)

    Total: 12,870 entries
    """

    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step41-LR-oblique-pairing-left-only.txt',
            '4182000041820000',
            linecount=12870)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Left
            'x', parent_state[45], 'x', 'x',
            'x', 'x', 'x', parent_state[53],
            parent_state[56], 'x', 'x', 'x',
            'x', 'x', parent_state[64], 'x',

            # Front
            'x', parent_state[81], 'x', 'x',
            'x', 'x', 'x', parent_state[89],
            parent_state[92], 'x', 'x', 'x',
            'x', 'x', parent_state[100], 'x',

            # Right
            'x', parent_state[117], 'x', 'x',
            'x', 'x', 'x', parent_state[125],
            parent_state[128], 'x', 'x', 'x',
            'x', 'x', parent_state[136], 'x',

            # Back
            'x', parent_state[153], 'x', 'x',
            'x', 'x', 'x', parent_state[161],
            parent_state[164], 'x', 'x', 'x',
            'x', 'x', parent_state[172], 'x',
        ]

        result = ['1' if x in ('L', 'R') else '0' for x in result]
        result = ''.join(result)

        # Convert to hex
        return self.hex_format % int(result, 2)


class LookupTable666LRObliqueEdgePairingRightOnly(LookupTable):
    """
    lookup-table-6x6x6-step42-LR-oblique-pairing-right-only.txt
    ===========================================================
    1 steps has 3 entries (0 percent, 0.00x previous step)
    2 steps has 29 entries (0 percent, 9.67x previous step)
    3 steps has 238 entries (1 percent, 8.21x previous step)
    4 steps has 742 entries (5 percent, 3.12x previous step)
    5 steps has 1,836 entries (14 percent, 2.47x previous step)
    6 steps has 4,405 entries (34 percent, 2.40x previous step)
    7 steps has 3,774 entries (29 percent, 0.86x previous step)
    8 steps has 1,721 entries (13 percent, 0.46x previous step)
    9 steps has 122 entries (0 percent, 0.07x previous step)

    Total: 12,870 entries
    """

    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step42-LR-oblique-pairing-right-only.txt',
            '2814000028140000',
            linecount=12870)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Left
            'x', 'x', parent_state[46], 'x',
             parent_state[50], 'x', 'x', 'x',
            'x', 'x', 'x', parent_state[59],
            'x', parent_state[63], 'x', 'x',

            # Front
            'x', 'x', parent_state[82], 'x',
            parent_state[86], 'x', 'x', 'x',
            'x', 'x', 'x', parent_state[95],
            'x', parent_state[99], 'x', 'x',

            # Right
            'x', 'x', parent_state[118], 'x',
            parent_state[122], 'x', 'x', 'x',
            'x', 'x', 'x', parent_state[131],
            'x', parent_state[135], 'x', 'x',

            # Back
            'x', 'x', parent_state[154], 'x',
            parent_state[158], 'x', 'x', 'x',
            'x', 'x', 'x', parent_state[167],
            'x', parent_state[171], 'x', 'x',
        ]

        result = ['1' if x in ('L', 'R') else '0' for x in result]
        result = ''.join(result)

        # Convert to hex
        return self.hex_format % int(result, 2)


class LookupTableIDA666LRObliqueEdgePairing(LookupTableIDA):
    """
    lookup-table-6x6x6-step40-LR-oblique-pairing.txt
    ================================================
    1 steps has 3 entries (0 percent, 0.00x previous step)
    2 steps has 29 entries (0 percent, 9.67x previous step)
    3 steps has 286 entries (0 percent, 9.86x previous step)
    4 steps has 2052 entries (0 percent, 7.17x previous step)
    5 steps has 16348 entries (1 percent, 7.97x previous step)
    6 steps has 127859 entries (12 percent, 7.82x previous step)
    7 steps has 844248 entries (85 percent, 6.60x previous step)

    Total: 990825 entries
    """

    def __init__(self, parent):
        LookupTableIDA.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step40-LR-oblique-pairing.txt',
            '6996000069960000',
            moves_6x6x6,

            # These would break up the staged UD inner x-centers
            ("3Rw", "3Rw'", "3Lw", "3Lw'", "3Fw", "3Fw'", "3Bw", "3Bw'", # do not mess up UD x-centers
             "Rw", "Rw'", "Lw", "Lw'", "Fw", "Fw'", "Bw", "Bw'",         # do not mess up UD oblique pairs
             "3Uw", "3Uw'", "3Dw", "3Dw'", "3Fw", "3Fw'", "3Bw", "3Bw'", # do not mess up LR x-centers
             "3Lw", "3Lw'", "3Lw2",        # can skip these for 6x6x6 cubes
             "3Dw", "3Dw'", "3Dw2",
             "3Bw", "3Bw'", "3Bw2"),

            # prune tables
            (parent.lt_LR_oblique_edge_pairing_left_only,
             parent.lt_LR_oblique_edge_pairing_right_only),
            linecount=990825)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Left
            'x', parent_state[45], parent_state[46], 'x',
            parent_state[50], 'x', 'x', parent_state[53],
            parent_state[56], 'x', 'x', parent_state[59],
            'x', parent_state[63], parent_state[64], 'x',

            # Front
            'x', parent_state[81], parent_state[82], 'x',
            parent_state[86], 'x', 'x', parent_state[89],
            parent_state[92], 'x', 'x', parent_state[95],
            'x', parent_state[99], parent_state[100], 'x',

            # Right
            'x', parent_state[117], parent_state[118], 'x',
            parent_state[122], 'x', 'x', parent_state[125],
            parent_state[128], 'x', 'x', parent_state[131],
            'x', parent_state[135], parent_state[136], 'x',

            # Back
            'x', parent_state[153], parent_state[154], 'x',
            parent_state[158], 'x', 'x', parent_state[161],
            parent_state[164], 'x', 'x', parent_state[167],
            'x', parent_state[171], parent_state[172], 'x',
        ]

        result = ['1' if x in ('L', 'R') else '0' for x in result]
        result = ''.join(result)

        # Convert to hex
        return self.hex_format % int(result, 2)


class LookupTable666UDInnerXCenterAndObliqueEdges(LookupTable):
    """
    This will solve the UD inner x-centers and pair the UD oblique edges.

    lookup-table-6x6x6-step50-UD-solve-inner-x-center-and-oblique-edges.txt
    =======================================================================
    1 steps has 350 entries (0 percent, 0.00x previous step)
    2 steps has 1,358 entries (0 percent, 3.88x previous step)
    3 steps has 5,148 entries (1 percent, 3.79x previous step)
    4 steps has 21,684 entries (6 percent, 4.21x previous step)
    5 steps has 75,104 entries (21 percent, 3.46x previous step)
    6 steps has 134,420 entries (39 percent, 1.79x previous step)
    7 steps has 91,784 entries (26 percent, 0.68x previous step)
    8 steps has 13,152 entries (3 percent, 0.14x previous step)

    Total: 343,000 entries
    """

    # TODO this pairs and solves the UD oblique edges, we need to build
    # the 70 tables that let them go anywhere on sides U or D
    # These are building now
    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step50-UD-solve-inner-x-center-and-oblique-edges.txt',
            ('xDDxDUUDDUUDxDDxxUUxUDDUUDDUxUUx',
             'xDDxDUUDDUUDxUUxxDDxUDDUUDDUxUUx',
             'xDDxDUUDDUUDxUUxxUUxDDDUDDDUxUUx',
             'xDDxDUUDDUUDxUUxxUUxUDDDUDDDxUUx',
             'xDDxDUUDDUUDxUUxxUUxUDDUUDDUxDDx',
             'xDDxDUUUDUUUxDDxxDDxUDDUUDDUxUUx',
             'xDDxDUUUDUUUxDDxxUUxDDDUDDDUxUUx',
             'xDDxDUUUDUUUxDDxxUUxUDDDUDDDxUUx',
             'xDDxDUUUDUUUxDDxxUUxUDDUUDDUxDDx',
             'xDDxDUUUDUUUxUUxxDDxDDDUDDDUxUUx',
             'xDDxDUUUDUUUxUUxxDDxUDDDUDDDxUUx',
             'xDDxDUUUDUUUxUUxxDDxUDDUUDDUxDDx',
             'xDDxDUUUDUUUxUUxxUUxDDDDDDDDxUUx',
             'xDDxDUUUDUUUxUUxxUUxDDDUDDDUxDDx',
             'xDDxDUUUDUUUxUUxxUUxUDDDUDDDxDDx',
             'xDDxUUUDUUUDxDDxxDDxUDDUUDDUxUUx',
             'xDDxUUUDUUUDxDDxxUUxDDDUDDDUxUUx',
             'xDDxUUUDUUUDxDDxxUUxUDDDUDDDxUUx',
             'xDDxUUUDUUUDxDDxxUUxUDDUUDDUxDDx',
             'xDDxUUUDUUUDxUUxxDDxDDDUDDDUxUUx',
             'xDDxUUUDUUUDxUUxxDDxUDDDUDDDxUUx',
             'xDDxUUUDUUUDxUUxxDDxUDDUUDDUxDDx',
             'xDDxUUUDUUUDxUUxxUUxDDDDDDDDxUUx',
             'xDDxUUUDUUUDxUUxxUUxDDDUDDDUxDDx',
             'xDDxUUUDUUUDxUUxxUUxUDDDUDDDxDDx',
             'xDDxUUUUUUUUxDDxxDDxDDDUDDDUxUUx',
             'xDDxUUUUUUUUxDDxxDDxUDDDUDDDxUUx',
             'xDDxUUUUUUUUxDDxxDDxUDDUUDDUxDDx',
             'xDDxUUUUUUUUxDDxxUUxDDDDDDDDxUUx',
             'xDDxUUUUUUUUxDDxxUUxDDDUDDDUxDDx',
             'xDDxUUUUUUUUxDDxxUUxUDDDUDDDxDDx',
             'xDDxUUUUUUUUxUUxxDDxDDDDDDDDxUUx',
             'xDDxUUUUUUUUxUUxxDDxDDDUDDDUxDDx',
             'xDDxUUUUUUUUxUUxxDDxUDDDUDDDxDDx',
             'xDDxUUUUUUUUxUUxxUUxDDDDDDDDxDDx',
             'xUUxDUUDDUUDxDDxxDDxUDDUUDDUxUUx',
             'xUUxDUUDDUUDxDDxxUUxDDDUDDDUxUUx',
             'xUUxDUUDDUUDxDDxxUUxUDDDUDDDxUUx',
             'xUUxDUUDDUUDxDDxxUUxUDDUUDDUxDDx',
             'xUUxDUUDDUUDxUUxxDDxDDDUDDDUxUUx',
             'xUUxDUUDDUUDxUUxxDDxUDDDUDDDxUUx',
             'xUUxDUUDDUUDxUUxxDDxUDDUUDDUxDDx',
             'xUUxDUUDDUUDxUUxxUUxDDDDDDDDxUUx',
             'xUUxDUUDDUUDxUUxxUUxDDDUDDDUxDDx',
             'xUUxDUUDDUUDxUUxxUUxUDDDUDDDxDDx',
             'xUUxDUUUDUUUxDDxxDDxDDDUDDDUxUUx',
             'xUUxDUUUDUUUxDDxxDDxUDDDUDDDxUUx',
             'xUUxDUUUDUUUxDDxxDDxUDDUUDDUxDDx',
             'xUUxDUUUDUUUxDDxxUUxDDDDDDDDxUUx',
             'xUUxDUUUDUUUxDDxxUUxDDDUDDDUxDDx',
             'xUUxDUUUDUUUxDDxxUUxUDDDUDDDxDDx',
             'xUUxDUUUDUUUxUUxxDDxDDDDDDDDxUUx',
             'xUUxDUUUDUUUxUUxxDDxDDDUDDDUxDDx',
             'xUUxDUUUDUUUxUUxxDDxUDDDUDDDxDDx',
             'xUUxDUUUDUUUxUUxxUUxDDDDDDDDxDDx',
             'xUUxUUUDUUUDxDDxxDDxDDDUDDDUxUUx',
             'xUUxUUUDUUUDxDDxxDDxUDDDUDDDxUUx',
             'xUUxUUUDUUUDxDDxxDDxUDDUUDDUxDDx',
             'xUUxUUUDUUUDxDDxxUUxDDDDDDDDxUUx',
             'xUUxUUUDUUUDxDDxxUUxDDDUDDDUxDDx',
             'xUUxUUUDUUUDxDDxxUUxUDDDUDDDxDDx',
             'xUUxUUUDUUUDxUUxxDDxDDDDDDDDxUUx',
             'xUUxUUUDUUUDxUUxxDDxDDDUDDDUxDDx',
             'xUUxUUUDUUUDxUUxxDDxUDDDUDDDxDDx',
             'xUUxUUUDUUUDxUUxxUUxDDDDDDDDxDDx',
             'xUUxUUUUUUUUxDDxxDDxDDDDDDDDxUUx',
             'xUUxUUUUUUUUxDDxxDDxDDDUDDDUxDDx',
             'xUUxUUUUUUUUxDDxxDDxUDDDUDDDxDDx',
             'xUUxUUUUUUUUxDDxxUUxDDDDDDDDxDDx',
             'xUUxUUUUUUUUxUUxxDDxDDDDDDDDxDDx'),
            linecount=343000)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Upper
            'x', parent_state[9], parent_state[10], 'x',
            parent_state[14], parent_state[15], parent_state[16], parent_state[17],
            parent_state[20], parent_state[21], parent_state[22], parent_state[23],
            'x', parent_state[27], parent_state[28], 'x',

            # Down
            'x', parent_state[189], parent_state[190], 'x',
            parent_state[194], parent_state[195], parent_state[196], parent_state[197],
            parent_state[200], parent_state[201], parent_state[202], parent_state[203],
            'x', parent_state[207], parent_state[208], 'x'
        ]

        result = ''.join(result)
        return result


class LookupTable666LRInnerXCenterAndObliqueEdges(LookupTable):
    """
    lookup-table-6x6x6-step61-LR-solve-inner-x-center-and-oblique-edges.txt
    ========================================================================
    1 steps has 140 entries (0 percent, 0.00x previous step)
    2 steps has 420 entries (0 percent, 3.00x previous step)
    3 steps has 1390 entries (0 percent, 3.31x previous step)
    4 steps has 5,794 entries (1 percent, 4.17x previous step)
    5 steps has 17,606 entries (5 percent, 3.04x previous step)
    6 steps has 51,580 entries (15 percent, 2.93x previous step)
    7 steps has 102,466 entries (29 percent, 1.99x previous step)
    8 steps has 114,764 entries (33 percent, 1.12x previous step)
    9 steps has 45,068 entries (13 percent, 0.39x previous step)
    10 steps has 3,656 entries (1 percent, 0.08x previous step)
    11 steps has 116 entries (0 percent, 0.03x previous step)

    Total: 343,000 entries
    """

    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step61-LR-solve-inner-x-center-and-oblique-edges.txt',
            ('xLLxLLLLLLLLxLLxxxxxxxxxxxxxxxxxxRRxRRRRRRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxRRxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxRRxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxRRxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxRRxxxxxxxxxxxxxxxxxxLLxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxRRxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxRRxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xLLxRLLRRLLRxRRxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxLLxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxLLxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxLLxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxLLxxxxxxxxxxxxxxxxxxRRxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLLLLLLxRRxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxLLxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxLLxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxLLLRLLLRxRRxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxRRxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxLLxRRRRRRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxRRxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxLLxxxxxxxxxxxxxxxxxxRRxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxLLxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLLRLLLxRRxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxLLxLRRLLRRLxRRxxxxxxxxxxxxxxxxx',
             'xRRxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxLLxLRRRLRRRxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxLLxRRRLRRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLRRLLRxLLxxxxxxxxxxxxxxxxxxRRxLRRLLRRLxLLxxxxxxxxxxxxxxxxx',
             'xRRxRLLRRLLRxRRxxxxxxxxxxxxxxxxxxLLxLRRLLRRLxLLxxxxxxxxxxxxxxxxx'),
            linecount=343000)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Left
            'x', parent_state[45], parent_state[46], 'x',
            parent_state[50], parent_state[51], parent_state[52], parent_state[53],
            parent_state[56], parent_state[57], parent_state[58], parent_state[59],
            'x', parent_state[63], parent_state[64], 'x',

            # Front
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',

            # Right
            'x', parent_state[117], parent_state[118], 'x',
            parent_state[122], parent_state[123], parent_state[124], parent_state[125],
            parent_state[128], parent_state[129], parent_state[130], parent_state[131],
            'x', parent_state[135], parent_state[136], 'x',

            # Back
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
        ]

        result = ''.join(result)
        return result


class LookupTable666FBInnerXCenterAndObliqueEdges(LookupTable):
    """
    lookup-table-6x6x6-step62-FB-solve-inner-x-center-and-oblique-edges.txt
    ========================================================================
    1 steps has 140 entries (0 percent, 0.00x previous step)
    2 steps has 420 entries (0 percent, 3.00x previous step)
    3 steps has 1,390 entries (0 percent, 3.31x previous step)
    4 steps has 5,794 entries (1 percent, 4.17x previous step)
    5 steps has 17,606 entries (5 percent, 3.04x previous step)
    6 steps has 51,580 entries (15 percent, 2.93x previous step)
    7 steps has 102,466 entries (29 percent, 1.99x previous step)
    8 steps has 114,764 entries (33 percent, 1.12x previous step)
    9 steps has 45,068 entries (13 percent, 0.39x previous step)
    10 steps has 3,656 entries (1 percent, 0.08x previous step)
    11 steps has 116 entries (0 percent, 0.03x previous step)

    Total: 343,000 entries
    """

    def __init__(self, parent):
        LookupTable.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step62-FB-solve-inner-x-center-and-oblique-edges.txt',
            ('xxxxxxxxxxxxxxxxxBBxBFFBBFFBxBBxxxxxxxxxxxxxxxxxxFFxFBBFFBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxFFxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxFFxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxFFxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxFFxxxxxxxxxxxxxxxxxxBBxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxFFxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxFFxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxBBxFFFFFFFFxFFxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxBBxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxBBxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxBBxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxBBxxxxxxxxxxxxxxxxxxFFxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFBBFFBxFFxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxBBxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxBBxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxBFFFBFFFxFFxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxFFx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxBBxFBBFFBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxFFxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxBBxxxxxxxxxxxxxxxxxxFFxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxBBxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFBFFFBxFFxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxBBxBBBBBBBBxFFx',
             'xxxxxxxxxxxxxxxxxFFxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxBBxBBBFBBBFxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxBBxFBBBFBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFFFFFFxBBxxxxxxxxxxxxxxxxxxFFxBBBBBBBBxBBx',
             'xxxxxxxxxxxxxxxxxFFxFFFFFFFFxFFxxxxxxxxxxxxxxxxxxBBxBBBBBBBBxBBx'),
            linecount=343000)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Left
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',

            # Front
            'x', parent_state[81], parent_state[82], 'x',
            parent_state[86], parent_state[87], parent_state[88], parent_state[89],
            parent_state[92], parent_state[93], parent_state[94], parent_state[95],
            'x', parent_state[99], parent_state[100], 'x',

            # Right
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',
            'x', 'x', 'x', 'x',

            # Back
            'x', parent_state[153], parent_state[154], 'x',
            parent_state[158], parent_state[159], parent_state[160], parent_state[161],
            parent_state[164], parent_state[165], parent_state[166], parent_state[167],
            'x', parent_state[171], parent_state[172], 'x',
        ]

        result = ''.join(result)
        return result


class LookupTableIDA666LFRBInnerXCenterAndObliqueEdges(LookupTableIDA):
    """
    lookup-table-6x6x6-step60-LFRB-solve-inner-x-center-and-oblique-edges.txt
    ==========================================================================
    1 steps has 9,800 entries (0 percent, 0.00x previous step)
    2 steps has 58,800 entries (0 percent, 6.00x previous step)
    3 steps has 390,600 entries (2 percent, 6.64x previous step)
    4 steps has 2,117,480 entries (15 percent, 5.42x previous step)
    5 steps has 11,105,284 entries (81 percent, 5.24x previous step)

    Total: 13,681,964 entries
    """

    def __init__(self, parent):
        LookupTableIDA.__init__(
            self,
            parent,
            'lookup-table-6x6x6-step60-LFRB-solve-inner-x-center-and-oblique-edges.txt',

            # There are 4900 of them so they are at the bottom of the file
            lt_LFRB_solve_inner_x_centers_and_oblique_edges_state_targets,

            moves_6x6x6,

            ("3Rw", "3Rw'", "3Lw", "3Lw'", "3Fw", "3Fw'", "3Bw", "3Bw'", "3Uw", "3Uw'", "3Dw", "3Dw'", # do not mess up staged centers
             "Rw", "Rw'", "Lw", "Lw'", "Fw", "Fw'", "Bw", "Bw'", "Uw", "Uw'", "Dw", "Dw'",             # do not mess up staged centers
             "3Rw2", "3Lw2", "3Fw2", "3Bw2", "Rw2", "Lw2", "Fw2", "Bw2",                               # do not mess up solved UD
             "3Lw", "3Lw'", "3Lw2",        # can skip these for 6x6x6 cubes
             "3Dw", "3Dw'", "3Dw2",
             "3Bw", "3Bw'", "3Bw2"),

            # prune tables
            (parent.lt_LR_solve_inner_x_centers_and_oblique_edges,
             parent.lt_FB_solve_inner_x_centers_and_oblique_edges,),
            linecount=13681964)

    def state(self):
        parent_state = self.parent.state

        result = [
            # Left
            'x', parent_state[45], parent_state[46], 'x',
            parent_state[50], parent_state[51], parent_state[52], parent_state[53],
            parent_state[56], parent_state[57], parent_state[58], parent_state[59],
            'x', parent_state[63], parent_state[64], 'x',

            # Front
            'x', parent_state[81], parent_state[82], 'x',
            parent_state[86], parent_state[87], parent_state[88], parent_state[89],
            parent_state[92], parent_state[93], parent_state[94], parent_state[95],
            'x', parent_state[99], parent_state[100], 'x',

            # Right
            'x', parent_state[117], parent_state[118], 'x',
            parent_state[122], parent_state[123], parent_state[124], parent_state[125],
            parent_state[128], parent_state[129], parent_state[130], parent_state[131],
            'x', parent_state[135], parent_state[136], 'x',

            # Back
            'x', parent_state[153], parent_state[154], 'x',
            parent_state[158], parent_state[159], parent_state[160], parent_state[161],
            parent_state[164], parent_state[165], parent_state[166], parent_state[167],
            'x', parent_state[171], parent_state[172], 'x'
        ]

        result = ''.join(result)
        return result


class RubiksCube666(RubiksCubeNNNEvenEdges):
    """
    6x6x6 strategy
    - stage UD centers to sides U or D (use IDA)
    - stage LR centers to sides L or R...this in turn stages FB centers to sides F or B
    - solve all centers (use IDA)
    - pair edges
    - solve as 3x3x3

    Inheritance model
    -----------------
            RubiksCube
                |
        RubiksCubeNNNEvenEdges
           /            \
    RubiksCubeNNNEven RubiksCube666
    """

    def sanity_check(self):
        edge_orbit_0 = (2, 5, 12, 30, 35, 32, 25, 7,
                        38, 41, 48, 66, 71, 68, 61, 43,
                        74, 77, 84, 102, 107, 104, 97, 79,
                        110, 113, 120, 138, 143, 140, 133, 115,
                        146, 149, 156, 174, 179, 176, 169, 151,
                        182, 185, 192, 210, 215, 212, 205, 187)

        edge_orbit_1 = (3, 4, 18, 24, 34, 33, 19, 13,
                        39, 40, 54, 60, 70, 69, 55, 49,
                        75, 76, 90, 96, 106, 105, 91, 85,
                        111, 112, 126, 132, 142, 141, 127, 121,
                        147, 148, 162, 168, 178, 177, 163, 157,
                        183, 184, 198, 204, 214, 213, 199, 193)

        corners = (1, 6, 31, 36,
                   37, 42, 67, 72,
                   73, 78, 103, 108,
                   109, 114, 139, 144,
                   145, 150, 175, 180,
                   181, 186, 211, 216)

        left_oblique_edge = (9, 17, 28, 20,
                             45, 53, 64, 56,
                             81, 89, 100, 92,
                             117, 125, 136, 128,
                             153, 161, 172, 164,
                             189, 197, 208, 200)

        right_oblique_edge = (10, 23, 27, 14,
                              46, 59, 63, 50,
                              82, 95, 99, 86,
                              118, 131, 135, 122,
                              154, 167, 171, 158,
                              190, 203, 207, 194)

        outside_x_centers = (8, 11, 26, 29,
                             44, 47, 62, 65,
                             80, 83, 98, 101,
                             116, 119, 134, 137,
                             152, 155, 170, 173,
                             188, 191, 206, 209)

        inside_x_centers = (15, 16, 21, 22,
                            51, 52, 57, 58,
                            87, 88, 93, 94,
                            123, 124, 129, 130,
                            159, 160, 165, 166,
                            195, 196, 201, 202)

        self._sanity_check('edge-orbit-0', edge_orbit_0, 8)
        self._sanity_check('edge-orbit-1', edge_orbit_1, 8)
        self._sanity_check('corners', corners, 4)
        self._sanity_check('left-oblique', left_oblique_edge, 4)
        self._sanity_check('right-oblique', right_oblique_edge, 4)
        self._sanity_check('outside x-center', outside_x_centers, 4)
        self._sanity_check('inside x-center', inside_x_centers, 4)

    def lt_init(self):
        if self.lt_init_called:
            return
        self.lt_init_called = True

        self.lt_UD_inner_x_centers_stage = LookupTable666UDInnerXCenterStage(self)

        self.lt_UD_oblique_edge_pairing_left_only = LookupTable666UDObliqueEdgePairingLeftOnly(self)
        self.lt_UD_oblique_edge_pairing_right_only = LookupTable666UDObliqueEdgePairingRightOnly(self)
        self.lt_UD_oblique_edge_pairing = LookupTableIDA666UDObliqueEdgePairing(self)

        self.lt_LR_inner_x_centers = LookupTable666LRInnerXCenters(self)

        self.lt_LR_oblique_edge_pairing_left_only = LookupTable666LRObliqueEdgePairingLeftOnly(self)
        self.lt_LR_oblique_edge_pairing_right_only = LookupTable666LRObliqueEdgePairingRightOnly(self)
        self.lt_LR_oblique_edge_pairing = LookupTableIDA666LRObliqueEdgePairing(self)

        self.lt_UD_solve_inner_x_centers_and_oblique_edges = LookupTable666UDInnerXCenterAndObliqueEdges(self)

        self.lt_LR_solve_inner_x_centers_and_oblique_edges = LookupTable666LRInnerXCenterAndObliqueEdges(self)
        self.lt_FB_solve_inner_x_centers_and_oblique_edges = LookupTable666FBInnerXCenterAndObliqueEdges(self)
        self.lt_LFRB_solve_inner_x_centers_and_oblique_edges = LookupTableIDA666LFRBInnerXCenterAndObliqueEdges(self)

    def populate_fake_555_for_ULFRBD_solve(self, fake_555):

        for x in range(1, 151):
            fake_555.state[x] = 'x'

        # Upper
        fake_555.state[7] = self.state[8]
        fake_555.state[8] = self.state[9]
        fake_555.state[9] = self.state[11]
        fake_555.state[12] = self.state[14]
        fake_555.state[13] = self.state[15]
        fake_555.state[14] = self.state[17]
        fake_555.state[17] = self.state[26]
        fake_555.state[18] = self.state[27]
        fake_555.state[19] = self.state[29]

        # Left
        fake_555.state[32] = self.state[44]
        fake_555.state[33] = self.state[45]
        fake_555.state[34] = self.state[47]
        fake_555.state[37] = self.state[50]
        fake_555.state[38] = self.state[51]
        fake_555.state[39] = self.state[53]
        fake_555.state[42] = self.state[62]
        fake_555.state[43] = self.state[63]
        fake_555.state[44] = self.state[65]

        # Front
        fake_555.state[57] = self.state[80]
        fake_555.state[58] = self.state[81]
        fake_555.state[59] = self.state[83]
        fake_555.state[62] = self.state[86]
        fake_555.state[63] = self.state[87]
        fake_555.state[64] = self.state[89]
        fake_555.state[67] = self.state[98]
        fake_555.state[68] = self.state[99]
        fake_555.state[69] = self.state[101]

        # Right
        fake_555.state[82] = self.state[116]
        fake_555.state[83] = self.state[117]
        fake_555.state[84] = self.state[119]
        fake_555.state[87] = self.state[122]
        fake_555.state[88] = self.state[123]
        fake_555.state[89] = self.state[125]
        fake_555.state[92] = self.state[134]
        fake_555.state[93] = self.state[135]
        fake_555.state[94] = self.state[137]

        # Back
        fake_555.state[107] = self.state[152]
        fake_555.state[108] = self.state[153]
        fake_555.state[109] = self.state[155]
        fake_555.state[112] = self.state[158]
        fake_555.state[113] = self.state[159]
        fake_555.state[114] = self.state[161]
        fake_555.state[117] = self.state[170]
        fake_555.state[118] = self.state[171]
        fake_555.state[119] = self.state[173]

        # Down
        fake_555.state[132] = self.state[188]
        fake_555.state[133] = self.state[189]
        fake_555.state[134] = self.state[191]
        fake_555.state[137] = self.state[194]
        fake_555.state[138] = self.state[195]
        fake_555.state[139] = self.state[197]
        fake_555.state[142] = self.state[206]
        fake_555.state[143] = self.state[207]
        fake_555.state[144] = self.state[209]
        fake_555.sanity_check()

    def populate_fake_555_for_ULFRBD_stage(self, fake_555):
        """
        This will be called when the inner x-centers and oblique edges are staged
        so we can set those to be the name of the side they are on.  We only need
        to bring in the outer x-centers from the 6x6x6.
        """

        for x in range(1, 151):
            fake_555.state[x] = 'x'

        # Upper
        fake_555.state[7] = self.state[8]
        fake_555.state[8] = 'U'
        fake_555.state[9] = self.state[11]
        fake_555.state[12] = 'U'
        fake_555.state[13] = 'U'
        fake_555.state[14] = 'U'
        fake_555.state[17] = self.state[26]
        fake_555.state[18] = 'U'
        fake_555.state[19] = self.state[29]

        # Left
        fake_555.state[32] = self.state[44]
        fake_555.state[33] = 'L'
        fake_555.state[34] = self.state[47]
        fake_555.state[37] = 'L'
        fake_555.state[38] = 'L'
        fake_555.state[39] = 'L'
        fake_555.state[42] = self.state[62]
        fake_555.state[43] = 'L'
        fake_555.state[44] = self.state[65]

        # Front
        fake_555.state[57] = self.state[80]
        fake_555.state[58] = 'F'
        fake_555.state[59] = self.state[83]
        fake_555.state[62] = 'F'
        fake_555.state[63] = 'F'
        fake_555.state[64] = 'F'
        fake_555.state[67] = self.state[98]
        fake_555.state[68] = 'F'
        fake_555.state[69] = self.state[101]

        # Right
        fake_555.state[82] = self.state[116]
        fake_555.state[83] = 'R'
        fake_555.state[84] = self.state[119]
        fake_555.state[87] = 'R'
        fake_555.state[88] = 'R'
        fake_555.state[89] = 'R'
        fake_555.state[92] = self.state[134]
        fake_555.state[93] = 'R'
        fake_555.state[94] = self.state[137]

        # Back
        fake_555.state[107] = self.state[152]
        fake_555.state[108] = 'B'
        fake_555.state[109] = self.state[155]
        fake_555.state[112] = 'B'
        fake_555.state[113] = 'B'
        fake_555.state[114] = 'B'
        fake_555.state[117] = self.state[170]
        fake_555.state[118] = 'B'
        fake_555.state[119] = self.state[173]

        # Down
        fake_555.state[132] = self.state[188]
        fake_555.state[133] = 'D'
        fake_555.state[134] = self.state[191]
        fake_555.state[137] = 'D'
        fake_555.state[138] = 'D'
        fake_555.state[139] = 'D'
        fake_555.state[142] = self.state[206]
        fake_555.state[143] = 'D'
        fake_555.state[144] = self.state[209]
        fake_555.sanity_check()

    def group_centers_stage_UD(self):
        """
        Stage UD inner x-centers and oblique edges. The 7x7x7 uses this that is why it is in its own method.
        """
        self.lt_UD_inner_x_centers_stage.solve()
        log.info("UD inner x-centers staged, %d steps in" % self.get_solution_len_minus_rotates(self.solution))

        # Test the prune tables
        #self.lt_UD_oblique_edge_pairing_left_only.solve()
        #self.lt_UD_oblique_edge_pairing_right_only.solve()
        #self.print_cube()
        #sys.exit(0)

        self.lt_UD_oblique_edge_pairing.solve()
        log.info("UD oblique edges paired, %d steps in" % self.get_solution_len_minus_rotates(self.solution))
        self.print_cube()
        log.info("")
        log.info("")
        log.info("")
        log.info("")
        log.info("")

    def group_centers_guts(self, oblique_edges_only=False):
        self.lt_init()

        self.group_centers_stage_UD()

        # Stage LR which in turn stages FB
        self.lt_LR_inner_x_centers.solve()
        log.info("LR inner x-centers staged, %d steps in" % self.get_solution_len_minus_rotates(self.solution))
        self.print_cube()
        log.info("")
        log.info("")
        log.info("")
        log.info("")
        log.info("")

        # Test the prune tables
        #self.lt_LR_oblique_edge_pairing_left_only.solve()
        #self.lt_LR_oblique_edge_pairing_right_only.solve()
        #self.print_cube()
        #sys.exit(0)

        self.lt_LR_oblique_edge_pairing.solve()
        log.info("LR oblique edges staged, %d steps in" % self.get_solution_len_minus_rotates(self.solution))
        self.print_cube()
        log.info("")
        log.info("")
        log.info("")
        log.info("")
        log.info("")
        # At this point all of the inner x-centers and oblique edges are staged

        solve_via_555 = True

        if solve_via_555 or oblique_edges_only:

            # Reduce the centers to 5x5x5 centers
            # - solve the UD inner x-centers and pair the UD oblique edges
            # - solve the LR inner x-centers and pair the LR oblique edges
            # - solve the FB inner x-centers and pair the FB oblique edges
            self.lt_UD_solve_inner_x_centers_and_oblique_edges.solve()
            log.info("UD inner x-center solved and oblique edges paired, %d steps in" % self.get_solution_len_minus_rotates(self.solution))
            self.print_cube()
            #log.info("kociemba: %s" % self.get_kociemba_string(True))
            log.info("")
            log.info("")
            log.info("")
            log.info("")
            log.info("")

            # Test the prune tables
            #self.lt_LR_solve_inner_x_centers_and_oblique_edges.solve()
            #self.lt_FB_solve_inner_x_centers_and_oblique_edges.solve()
            #self.print_cube()
            #sys.exit(0)

            self.lt_LFRB_solve_inner_x_centers_and_oblique_edges.solve()
            log.info("LFRB inner x-center and oblique edges paired, %d steps in" % self.get_solution_len_minus_rotates(self.solution))
            self.print_cube()
            log.info("")
            log.info("")
            log.info("")
            log.info("")
            log.info("")

            if oblique_edges_only:
                log.info("Took %d steps to resolve oblique edges" % self.get_solution_len_minus_rotates(self.solution))
                return

            # At this point the 6x6x6 centers have been reduced to 5x5x5 centers
            fake_555 = RubiksCube555(solved_5x5x5, 'URFDLB')
            fake_555.cpu_mode = self.cpu_mode
            fake_555.lt_init()
            self.populate_fake_555_for_ULFRBD_solve(fake_555)
            fake_555.group_centers_guts()

            for step in fake_555.solution:
                self.rotate(step)

        else:
            # For test cube FBDDDFFUDRFBBLFLLURLDLLUFBLRFDUFLBLLFBFLRRBBFDRRDUBUFRBUBRDLUBFDRLBBRLRUFLBRBDUDFFFDBLUDBBLRDFUUDLBBBRRDRUDLBLDFRUDLLFFUUBFBUUFDLRUDUDBRRBBUFFDRRRDBULRRURULFDBRRULDDRUUULBLLFDFRRFDURFFLDUUBRUFDRFUBLDFULFBFDDUDLBLLRBL
            # it takes 32 moves to get to this point.  If solve_via_555 is True we solve the 6x6x6
            # centers in 74 moves total...so this experiment needs to beat 74 moves.
            #
            # The code below gets all of the centers staged (via fake_555) in 53 moves.
            # Once all sides are staged there are 4 groups of 8!/(4!*4!) per side so to
            # solve ULF (which would also solve RBD) would be (8!/(4!*4!))^12 or
            # 13,841,287,201,000,000,000,000...that doesn't seem feasible.
            #
            # Solving just U (in turn solves D) would be (8!/(4!*4!))^4 or 24,010,000
            #
            # Solving L and F (in turn solves R and B) would be (8!/(4!*4!))^8 or
            # 576,480,100,000,000 but with two 24,010,000 prune tables for a ratio
            # of 0.000 000 041 6493128...it might be doable.
            #
            # The solve_via_555 code path can solve the centers in 74 moves while it takes
            # the non-solve_via_555 code 53 moves to get the centers staged.  That only
            # leaves us 21 moves to solve all of the centers...that doesn't sound like
            # much wiggle room and would take a lot of CPU to do the "solve L and R" IDA.
            #
            # This doesn't feel worth the effort to explore further.  I'll leave this bit
            # of code here for a rainy day though.

            fake_555 = RubiksCube555(solved_5x5x5, 'URFDLB')
            fake_555.cpu_mode = self.cpu_mode
            fake_555.lt_init()
            self.populate_fake_555_for_ULFRBD_stage(fake_555)
            fake_555.group_centers_stage_UD()
            fake_555.group_centers_stage_LR()

            for step in fake_555.solution:
                self.rotate(step)

            self.print_cube()
            log.info("Took %d steps to stage centers" % self.get_solution_len_minus_rotates(self.solution))
            sys.exit(0)

        log.info("Took %d steps to solve centers" % self.get_solution_len_minus_rotates(self.solution))

    def phase(self):
        if self._phase is None:
            self._phase = 'Stage UD centers'
            return self._phase

        if self._phase == 'Stage UD centers':
            if self.UD_centers_staged():
                self._phase = 'Stage LR centers'
            return self._phase

        if self._phase == 'Stage LR centers':
            if self.LR_centers_staged():
                self._phase = 'Solve Centers'

        if self._phase == 'Solve Centers':
            if self.centers_solved():
                self._phase = 'Pair Edges'

        if self._phase == 'Pair Edges':
            if not self.get_non_paired_edges():
                self._phase = 'Solve 3x3x3'

        return self._phase


lt_LFRB_solve_inner_x_centers_and_oblique_edges_state_targets = (
    'xLLxLLLLLLLLxLLxxBBxBFFBBFFBxBBxxRRxRRRRRRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxLLxxFFxFFFFFFFFxFFxxRRxRRRRRRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xLLxRLLRRLLRxRRxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxLLxxFFxFFFFFFFFxFFxxRRxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLLLLLLxRRxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxLLxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxLLLRLLLRxRRxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxBBxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFBBFFBxFFxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxBFFFBFFFxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxLLxRRRRRRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxBBxxRRxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFBFFFBxFFxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxLLxRRRRRRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxBBxxRRxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxRRxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxLLxRRRRRRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxRRxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxLLxxFFxFFFFFFFFxFFxxRRxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLLRLLLxRRxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxBBxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFBBFFBxFFxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxBFFFBFFFxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxRRxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxLRRRLRRRxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxLLxRRRLRRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxBBxxRRxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFBFFFBxFFxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxRRxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxLRRRLRRRxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxLLxRRRLRRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxBBxxRRxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxLLxLRRLLRRLxRRxxBBxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxLLxLRRRLRRRxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxLLxRRRLRRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxLLxxFFxFFFFFFFFxFFxxRRxLRRLLRRLxLLxxBBxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxRRxxBBxBFFBBFFBxBBxxLLxLRRLLRRLxLLxxFFxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxBBxFFFFFFFFxFFxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxBBxxLLxLRRLLRRLxLLxxFFxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFBBFFBxFFxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxBBxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxBFFFBFFFxFFxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxFFx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxBBxFBBFFBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxFFxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxBBxxLLxLRRLLRRLxLLxxFFxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFBFFFBxFFxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxBBxBBBBBBBBxFFx',
    'xRRxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxBBxBBBFBBBFxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxBBxFBBBFBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFFFFFFxBBxxLLxLRRLLRRLxLLxxFFxBBBBBBBBxBBx',
    'xRRxRLLRRLLRxRRxxFFxFFFFFFFFxFFxxLLxLRRLLRRLxLLxxBBxBBBBBBBBxBBx',
)
